var Q = require('q')
var Url = require('url')

var _sigFn

//Simple http get function
function _download_webpage(url){
    var deferred = Q.defer();

    var parsed_url = Url.parse(url)
    var options = {
        "method": "GET",
        "hostname": parsed_url.hostname,
        "port": null,
        "path": '/' + parsed_url.path
    };

    var http
    if(parsed_url.protocol.indexOf('https')!=-1)
        http = require('https')
    else
        http = require('http')

    var req = http.request(options, function (res) {
        var chunks = [];

        res.on("data", function (chunk) {
            chunks.push(chunk);
        });

        res.on("end", function () {
            var body = Buffer.concat(chunks);
            deferred.resolve(body.toString());
        });
    });

    req.on('error', function(err){
        deferred.reject(err);
    });

    req.end();

    return deferred.promise;
}


function _init(cipher){
    return Q().then(function(){
        return _download_webpage(cipher.url)
    }).then(function(code){
        cipher.code = code
        cipher._deferred.resolve()
    }).catch(cipher._deferred.reject)
}

function CipherExtract(url){
    this.url = url
    this._deferred = Q.defer()
    this._loadedPromise = this._deferred.promise
    _init(this)
}

/**
 * Get the signature extraction function name
 * @returns {Q.Promise<String>}
 */
CipherExtract.prototype.getSigFunctionName = function(){
    var self = this
    return self._loadedPromise.then(function(){
        var m = /\.sig\|\|([a-zA-Z0-9$]+)\(/g.exec(self.code)
        return m? m[1]:null
    })
}

/**
 * Get the signature extraction function
 * @returns {Q.Promise<function>}
 */
CipherExtract.prototype.getSigFunction = function(){
    var self = this

    if(_sigFn)
        return Q.when(_sigFn)

    return Q().then(function(){
        return self.getSigFunctionName()
    }).then(function(fnName){
        function _extractFnBody(){
            if(fnName){
                var re = new RegExp("(?:function\\\s+"+fnName+"|[{;,]\\\s*"+fnName+"\\\s*=\\\s*function|var\\\s+"+fnName+"\\\s*=\\\s*function)\\\s*\\\(([^)]*)\\\)\\\s*\\\{([^}]+)\\\}", 'gm')
                var m = re.exec(self.code)
                if(m)
                    return m[2]
            }
            return ''
        }

        function _extractUtilObject(){
            var utilObject = {
                name: null,
                functions: {}
            }
            var fnBody = _extractFnBody()
            if(!fnBody)
                throw new ReferenceError('Missing function body')
            var lines = fnBody.split(';')
            lines.forEach(function(line, index){
                if(index==0){
                    if(/[a-zA-Z]\.split\(["']*?\)/.exec(line))
                        return
                    else
                        console.log('WARNING : Unknown function extraction process')
                }
                if(index==lines.length-1){
                    if(/[a-zA-Z]\.join\(["']*?\)/.exec(line))
                        return
                    else
                        console.log('WARNING : Unknown function extraction process')
                }

                var m
                m = /([a-zA-Z0-9]+?)\.([a-zA-Z0-9]+?)\(.*?\)/.exec(line)
                if(m){
                    var name = m[1]
                    var fnName = m[2]
                    if(!utilObject['name'])
                        utilObject['name'] = name
                    else if(utilObject['name']!=name){
                        console.log('WARNING : Two or more util objects')
                        return
                    }
                    utilObject.functions[fnName] = true
                }
                else{
                    console.log('WARNING : unknown expression', line)
                }
            })

            for(var fnName in utilObject.functions){
                var re = new RegExp(fnName + '\\\s*?:\\\s*?function\\\(.*?\\\)\\\{(.*?)\\\}')
                var m = re.exec(self.code)
                if(m){
                    if(m[1].indexOf('var c=a[0];a[0]=a[b%a.length];a[b]=c')!=-1)
                        utilObject.functions[fnName] = function(a,b){var c = a[0]; a[0] = a[b%a.length]; a[b] = c; return a}
                    else if(m[1].indexOf('a.splice(0,b)')!=-1)
                        utilObject.functions[fnName] = function(a,b){a.splice(0, b); return a}
                    else if(m[1].indexOf('a.reverse()')!=-1)
                        utilObject.functions[fnName] = function(a){a.reverse(); return a}
                    else
                        console.log('WARNING : Unknown util function', fnName, ':', m[1])
                }
                else{
                    console.log('WARNING : Function', fnName, 'not found')
                }
            }
            return utilObject
        }

        function _getUtilFunction(stmt, utilObject){
            var m = /[a-zA-Z0-9]+?\.([a-zA-Z0-9]+?)\(.*?\)/.exec(stmt)
            return m && utilObject.functions[m[1]]? utilObject.functions[m[1]] : null
        }


        if(!fnName)
            throw new ReferenceError('Missing function name')

        var utilObject = _extractUtilObject()
        var lines = _extractFnBody().split(';')
        var processes = []
        lines.forEach(function(line, index){
            if(index==0){
                if(/[a-zA-Z]\.split\(["']*?\)/.exec(line))
                    processes.push(function(a){return a.split('');})
                else
                    console.log('WARNING : Unknown function extraction process')
            }
            else if(index==lines.length-1){
                if(/[a-zA-Z]\.join\(["']*?\)/.exec(line))
                    processes.push(function(a){return a.join('')})
                else
                    console.log('WARNING : Unknown function extraction process')
            }
            else{
                var m = /[a-zA-Z0-9]+?\.[a-zA-Z0-9]+?\((.*?)\)/.exec(line)
                if(m){
                    var fn = _getUtilFunction(line, utilObject)
                    if(fn){
                        var params = m[1].split(',')
                        if(params.length==2){
                            processes.push(function(a){return fn(a, parseInt(params[1]))})
                        }
                        else{
                            console.log('WARNING : Wrong arguments number for statment', line)
                        }
                    }
                    else{
                        console.log('WARNING : Function for statement', line, 'not found')
                    }
                }
                else{
                    console.log('WARNING : Unknown statement', line)
                }
            }
        })

        _sigFn = function(a){
            processes.forEach(function(proc){
                a = proc(a)
            })
            return a
        }
        return _sigFn
    })
}

module.exports = CipherExtract